currency_formats = {
    "eur": ("€ {}", "€ {}¢"),
    "sek": ("{} kr", "{} öre"),
    "nok": ("{} kr", "{} øre"),
    "usd": ("${}", "${}¢"),
    "gbp": ("£{}", "£{}p")
}

def currency_translation(value: float | str | None, currency, use_cent: bool = False) -> str:
    value = "-" if value is None else value
    if currency.lower() in currency_formats:
        format_string = currency_formats[currency.lower()]
        ret = format_string[0].format(value) if not use_cent else format_string[1].format(value)
    else:
        ret = f"{value} {currency}"
    return ret



from ....models.locale.enums.querytype import QueryType
from ..querytypes.querytypes import QUERYTYPES
from ..locale_model import Locale_Type
from ....models.locale.price.locale_price import LocalePrice
from ....models.locale.enums.price_type import PriceType
from ....models.locale.price.models.tiered_price import TieredPrice
from dataclasses import dataclass


@dataclass
class NO_Tensio(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Tiered,
            currency="NOK",
            _values=[
                TieredPrice(
                    upper_peak_limit=2,
                    value=126,
              ),
              TieredPrice(
                    upper_peak_limit=5,
                    value=225,
              ),
              TieredPrice(
                    upper_peak_limit=10,
                    value=385,
              ),
              TieredPrice(
                    upper_peak_limit=15,
                    value=567,
              ),
              TieredPrice(
                    upper_peak_limit=20,
                    value=749,
              ),
              TieredPrice(
                    upper_peak_limit=25,
                    value=933,
              ),
              TieredPrice(
                    upper_peak_limit=50,
                    value=1603,
              ),
              TieredPrice(
                    upper_peak_limit=75,
                    value=2516,
              ),
              TieredPrice(
                    upper_peak_limit=100,
                    value=3429,
              )
            ])

        """
        0 - 2	 126
        2 - 5	225
        5 - 10	 385
        10 - 15	 567
        15 - 20	 749
        20 - 25	 933
        25 - 50	1 603
        50 - 75	2 516
        75 - 100
        3 429
        100 - 150	4 953
        150 - 200	6 778
        200 - 300 	9 822
        300 - 400 	13 479
        400 - 500 	17 130
        0ver 500 	20 785
        """


@dataclass
class NO_LNett(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Tiered,
            currency="NOK",
            _values=[
                TieredPrice(
                    upper_peak_limit=5,
                    value=285,
              ),
              TieredPrice(
                    upper_peak_limit=10,
                    value=435,
              ),
              TieredPrice(
                    upper_peak_limit=15,
                    value=585,
              ),
              TieredPrice(
                    upper_peak_limit=20,
                    value=735,
              ),
              TieredPrice(
                    upper_peak_limit=25,
                    value=885,
              )
            ])

# docs: https://www.l-nett.no/nynettleie/slik-blir-ny-nettleie-og-pris


@dataclass
class NO_GlitreEnergi(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Tiered,
            currency="NOK",
            _values=[
                TieredPrice(
                    upper_peak_limit=2,
                    value=155,
              ),
              TieredPrice(
                    upper_peak_limit=5,
                    value=195,
              ),
              TieredPrice(
                    upper_peak_limit=10,
                    value=335,
              ),
              TieredPrice(
                    upper_peak_limit=15,
                    value=690,
              ),
              TieredPrice(
                    upper_peak_limit=20,
                    value=900,
              )
            ])


# docs: https://www.glitreenergi-nett.no/smart-nettleie/


@dataclass
class NO_AgderEnergi(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Tiered,
            currency="NOK",
            _values=[
                TieredPrice(
                    upper_peak_limit=2,
                    value=155,
              ),
              TieredPrice(
                    upper_peak_limit=5,
                    value=195,
              ),
              TieredPrice(
                    upper_peak_limit=10,
                    value=335,
              ),
              TieredPrice(
                    upper_peak_limit=15,
                    value=690,
              ),
              TieredPrice(
                    upper_peak_limit=20,
                    value=900,
              )
            ])

@dataclass
class NO_Elvia(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Tiered,
            currency="NOK",
            _values=[
                TieredPrice(
                    upper_peak_limit=2,
                    value=120,
              ),
              TieredPrice(
                    upper_peak_limit=5,
                    value=190,
              ),
              TieredPrice(
                    upper_peak_limit=10,
                    value=305,
              ),
              TieredPrice(
                    upper_peak_limit=15,
                    value=420,
              ),
              TieredPrice(
                    upper_peak_limit=20,
                    value=535,
              )
            ])

"""
0-2     120
2-5     190
5-10    305
10-15   420
15-20   535
"""

@dataclass
class NO_Lede(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Tiered,
            currency="NOK",
            _values=[
              TieredPrice(
                    upper_peak_limit=5,
                    value=262.5,
              ),
              TieredPrice(
                    upper_peak_limit=10,
                    value=448.75,
              ),
              TieredPrice(
                    upper_peak_limit=15,
                    value=632.5,
              ),
              TieredPrice(
                    upper_peak_limit=20,
                    value=818.75,
              ),
              TieredPrice(
                    upper_peak_limit=25,
                    value=1003.75,
              ),
              TieredPrice(
                    upper_peak_limit=50,
                    value=1560,
              ),
              TieredPrice(
                    upper_peak_limit=75,
                    value=2486.25,
              ),
              TieredPrice(
                    upper_peak_limit=100,
                    value=3412.5,
              ),
            ])

@dataclass
class NO_BKK(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Tiered,
            currency="NOK",
            _values=[
              TieredPrice(
                    upper_peak_limit=2,
                    value=145,
              ),
              TieredPrice(
                    upper_peak_limit=5,
                    value=240,
              ),
              TieredPrice(
                    upper_peak_limit=10,
                    value=400,
              ),
              TieredPrice(
                    upper_peak_limit=15,
                    value=570,
              ),
              TieredPrice(
                    upper_peak_limit=20,
                    value=735,
              ),
              TieredPrice(
                    upper_peak_limit=25,
                    value=900,
              )
            ])

@dataclass
class NO_Mellom(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Tiered,
            currency="NOK",
            _values=[
              TieredPrice(
                    upper_peak_limit=2,
                    value=243,
              ),
              TieredPrice(
                    upper_peak_limit=5,
                    value=364,
              ),
              TieredPrice(
                    upper_peak_limit=10,
                    value=607,
              ),
              TieredPrice(
                    upper_peak_limit=15,
                    value=801,
              ),
              TieredPrice(
                    upper_peak_limit=20,
                    value=1008,
              ),
              TieredPrice(
                    upper_peak_limit=25,
                    value=1263,
              ),
              TieredPrice(
                    upper_peak_limit=50,
                    value=1590,
              ),
              TieredPrice(
                    upper_peak_limit=999,
                    value=2125,
              )
            ])


@dataclass
class NO_AskerNett(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Tiered,
            currency="NOK",
            _values=[
                TieredPrice(
                    upper_peak_limit=2,
                    value=185,
                ),
                TieredPrice(
                    upper_peak_limit=5,
                    value=230,
                ),
                TieredPrice(
                    upper_peak_limit=10,
                    value=340,
                ),
                TieredPrice(
                    upper_peak_limit=15,
                    value=710,
                ),
                TieredPrice(
                    upper_peak_limit=20,
                    value=895,
                ),
                TieredPrice(
                    upper_peak_limit=25,
                    value=1130,
                ),
                TieredPrice(
                    upper_peak_limit=50,
                    value=1600,
                ),
                TieredPrice(
                    upper_peak_limit=75,
                    value=2540,
                ),
                TieredPrice(
                    upper_peak_limit=100,
                    value=3380,
                ),
                TieredPrice(
                    upper_peak_limit=999,
                    value=5400,
                ),
            ])
    # docs: https://askernett.no/ny-nettleiemodell-for-alle-fra-1-juli-2022/


@dataclass
class NO_Linja(Locale_Type):
    def __post_init__(self):
        self.observed_peak = QueryType.AverageOfThreeDays
        self.charged_peak = QueryType.AverageOfThreeDays
        self.query_model = QUERYTYPES[QueryType.AverageOfThreeDays]
        self.price = LocalePrice(
            price_type=PriceType.Tiered,
            currency="NOK",
            _values=[
                TieredPrice(
                    upper_peak_limit=2,
                    value=263,
                ),
                TieredPrice(
                    upper_peak_limit=5,
                    value=329,
                ),
                TieredPrice(
                    upper_peak_limit=10,
                    value=395,
                ),
                TieredPrice(
                    upper_peak_limit=15,
                    value=658,
                ),
                TieredPrice(
                    upper_peak_limit=20,
                    value=789,
                ),
                TieredPrice(
                    upper_peak_limit=25,
                    value=920,
                ),
                TieredPrice(
                    upper_peak_limit=50,
                    value=1315,
                ),
                TieredPrice(
                    upper_peak_limit=75,
                    value=1446,
                ),
                TieredPrice(
                    upper_peak_limit=100,
                    value=1578,
                ),
                TieredPrice(
                    upper_peak_limit=999,
                    value=1973,
                ),
            ])
        
        """
        Trinn 1	0-2	263
        Trinn 2	2-5	329
        Trinn 3	5-10	395
        Trinn 4	10-15	658
        Trinn 5	15-20	789
        Trinn 6	20-25	920
        Trinn 7	25-50	1 315
        Trinn 8	50-75	1 446
        Trinn 9	75-100	1 578
        Trinn 10	Over 100	1 973
        """
    #docs: https://www.morenett.no/informasjon/nettleie-privat






from .hoursbase import Hours
from datetime import datetime, timedelta


class RegularHours(Hours):
    def __init__(self, hub):
        self.hub = hub
        self.timer = None
        self.scheduler = None
        super().__init__(
            False, self.hub.options.nonhours, self.hub.options.cautionhours
        )

    @property
    def non_hours(self):
        return self.create_hours(self._non_hours)

    @non_hours.setter
    def non_hours(self, val):
        self._non_hours = val

    @property
    def caution_hours(self):
        return self.create_hours(self._caution_hours)

    @caution_hours.setter
    def caution_hours(self, val):
        self._caution_hours = val

    @property
    def is_initialized(self) -> bool:
        return True

    async def async_update_max_min(
        self,
        max_charge: float,
        session_energy: float | None = None,
        car_connected: bool = False,
        limiter: float = 0.0
    ):
        pass

    @property
    def absolute_top_price(self):
        return None

    @property
    def min_price(self):
        return None

    @property
    def future_hours(self) -> list: #type: ignore
        pass

    def create_hours(self, input:list = []) -> list[datetime]:
        hours = []
        for hour in input:
            if hour < datetime.now().hour:
                hours.append(datetime.now().replace(hour=hour, minute=0, second=0, microsecond=0) + timedelta(days=1))
            else:
                hours.append(datetime.now().replace(hour=hour, minute=0, second=0, microsecond=0))
        return hours


from __future__ import annotations
from dataclasses import dataclass, field
from .models.seasoned_price import SeasonedPrice
from .models.tiered_price import TieredPrice
from ..enums.price_type import PriceType

@dataclass
class LocalePrice:
    price_type: PriceType = field(default_factory=lambda: PriceType.Unset)
    value: float = 0.0
    _values: list[TieredPrice] | list[SeasonedPrice] | list = field(default_factory=lambda: [])
    currency: str = ""

    @property
    def is_active(self) -> bool:
        return self.price_type != PriceType.Unset

    def __post_init__(self):
        match self.price_type:
            case PriceType.Unset:
                pass
            case PriceType.Static:
                assert isinstance(self.value, (float, int))
                assert len(self.currency)
            case PriceType.Seasoned:
                assert isinstance(self._values[0], SeasonedPrice)
                assert len(self.currency)
            case PriceType.Tiered:
                assert isinstance(self._values[0], TieredPrice)
                assert len(self.currency)

    @property
    def price(self) -> float:
        match self.price_type:
            case PriceType.Static:
                return self.value
            case PriceType.Seasoned:
                return self._get_price_seasoned()
            case PriceType.Tiered:
                return self._get_price_tiered()
            case _:
                return 0.0

    def is_equal(self, other_currency: str) -> bool:
        """Use this method if necessary to test against the el-price currency"""
        return self.currency.lower == other_currency.lower

    def _get_price_seasoned(self) -> float:
        s: SeasonedPrice
        for s in [s for s in self._values if isinstance(s, SeasonedPrice)]:
            if s.validity.valid():
                return s.value
        return self.value

    def _get_price_tiered(self) -> float:
        for t in [s for s in self._values if isinstance(s, TieredPrice)]:
            pass
        return self.value



from __future__ import annotations
from typing import TYPE_CHECKING
from datetime import datetime, timedelta

from .models.permittance_type import PermittanceType
from .models.hour_type import HourType
from ..hoursselection_service_new.models.hour_price import HourPrice
import copy
if TYPE_CHECKING:
    from .hourselection_service import HourSelectionService
from .models.max_min_model import MaxMinModel


class MaxMinHourCap:
    def __init__(self) -> None:
        self._car_connected: bool = False
        self._allow_calculate_cap: bool = False
        self._current_cap: datetime = datetime.max

    @property
    def do_calculate_cap(self) -> bool:
        return self._allow_calculate_cap

    @property
    def car_connected(self) -> bool:
        return self._car_connected
    
    @car_connected.setter
    def car_connected(self, val: bool) -> None:
        if val != self._car_connected:
            self._allow_calculate_cap = True
        self._car_connected = val

    def cap(self, hours: list[HourPrice]) -> datetime:
        if self._allow_calculate_cap or min([x.dt for x in hours if not x.passed]) >= self._current_cap:
            current_hour = min([x.dt for x in hours if not x.passed])
            end_hour = datetime.max
            match current_hour.hour:
                case 9|10|11|12|13|14|15|16|17|18|19|20|21|22|23:
                    end_hour = (current_hour + timedelta(days=1)).replace(hour=9)
                case 0|1|2|3|4|5|6|7|8:
                    end_hour = (current_hour).replace(hour=18)
            self._current_cap = end_hour
            self._allow_calculate_cap = False
        return self._current_cap


class MaxMinCharge:
    def __init__(self, service: HourSelectionService, min_price: float | None) -> None:
        self.model = MaxMinModel(min_price=min_price)  # type: ignore
        self.hour_cap = MaxMinHourCap()
        self.parent = service
        self.active: bool = False
        self.overflow: bool = False
        self.current_peak: float | None = None

    @property
    def average_price(self) -> float | None:
        if not self.active:
            return None
        return self.model.caluclate_average_price(
            self.model.input_hours,
            self.total_charge,
            self.current_peak,
        )

    @property
    def original_average_price(self) -> float | None:
        return self.model.caluclate_average_price(
            self.parent.future_hours,
            self.original_total_charge,
            self.current_peak,
        )

    @property
    def total_charge(self) -> float | None:
        if not self.active:
            return None
        return self.model.calculate_total_charge(self.model.input_hours)

    @property
    def original_total_charge(self) -> float:
        return self.model.calculate_total_charge(self.parent.future_hours)

    def future_hours(self, dtmodel) -> list[HourPrice]:
        for hp in self.model.input_hours:
            hp.set_passed(dtmodel)
        ret = [hp for hp in self.model.input_hours if not hp.passed]
        return sorted(ret, key=lambda x: x.dt)

    async def async_update(
        self,
        avg24,
        peak,
        max_desired: float,
        session_energy: float | None = None,
        car_connected: bool = False,
        limiter: float = 0.0
    ) -> None:
        self.hour_cap.car_connected = car_connected
        if not car_connected:
            await self.async_setup(max_charge=peak)
        _session = session_energy or 0
        _desired = max_desired - _session
        _avg24 = round((avg24 / 1000), 1)
        self.model.expected_hourly_charge = max(peak - _avg24, 0)
        self.current_peak = peak
        if _desired >= self.original_total_charge:
            self.overflow = True
            self.get_hours()
            return
        else:
            self.overflow = False
            self.select_hours_for_charge(
                copy.deepcopy(self.parent.future_hours), _desired
            )            
            if self._price_is_too_similar(limiter):
                self.overflow = True
                self.get_hours()
                return

    def _price_is_too_similar(self, limiter: float) -> bool:
        original = getattr(self, "original_average_price",0)
        adjusted = getattr(self, "average_price",0)
        if original > 0 and adjusted > 0:
            return max(original,adjusted) - min(original,adjusted) < limiter
        return False

    def _cap_future_hours(self, hours: list[HourPrice]) -> list[HourPrice]:
        if self.hour_cap.do_calculate_cap:
            hours = [setattr(hour, 'permittance_type', PermittanceType.Regular) or hour for hour in hours]
        cap = self.hour_cap.cap(hours)
        maxmin_hours = [setattr(x, 'permittance_type', PermittanceType.MaxMin) or x for x in hours if x.dt <= cap]
        return maxmin_hours

    def select_hours_for_charge(
        self, hours: list[HourPrice], desired_charge: float
    ) -> None:
        _capped_hours = self._cap_future_hours(hours)
        _original_charge = self._get_charge_sum(_capped_hours)
        _total_charge: float = 0
        _desired: float = min([desired_charge, _original_charge])
        while _total_charge < _desired:
            _capped_hours.sort(key=lambda x: (x.price,x.dt))
            for hour in _capped_hours:
                if any([hour.passed, hour.permittance == 0, _total_charge >= _desired, hour.hour_type is HourType.AboveMax]):
                    hour.permittance = 0
                    continue
                _hour_charge = hour.permittance * self.model.expected_hourly_charge
                _perm = min(1,max((_desired - _total_charge) / self.model.expected_hourly_charge,0))
                _total_charge += _hour_charge * _perm
                hour.permittance = round(_perm, 2)
                if self._get_charge_sum(_capped_hours) <= desired_charge:
                    break
        self.model.input_hours = self._combine_hour_lists(hours, _capped_hours)

    def _combine_hour_lists(self, hours: list[HourPrice], capped_hours: list[HourPrice]) -> list[HourPrice]:
        capped_hours_dict = {x.dt: x for x in capped_hours}
        for i, hour in enumerate(hours):
            if hour.dt in capped_hours_dict:
                hours[i] = capped_hours_dict[hour.dt]
        return hours

    def _get_charge_sum(self, hours: list[HourPrice]) -> float:
        return sum(
            [
                hour.permittance * self.model.expected_hourly_charge
                for hour in hours
                if not hour.passed
            ]
        )

    def _set_expected_charge(self, desired, peak, avg24) -> float:
        return (desired - self.total_charge) / (peak - avg24)

    async def async_sum_charge(self, avg24, peak) -> float:
        total = 0
        for k in self.model.input_hours:
            total += (peak - avg24) * k.permittance
        return total

    async def async_setup(
        self,
        max_charge: float,
    ) -> None:
        if max_charge == 0:
            self.active = False
            return
        if not self.active:
            self.active = True
            self.get_hours()

    def get_hours(self) -> None:
        if self.active:
            self.model.input_hours = copy.deepcopy(self.parent.future_hours)
        # self.model.original_input_hours = copy.deepcopy(self.model.input_hours)



import logging
from typing import Tuple
from .topprice_type import TopPriceType
from statistics import mean

_LOGGER = logging.getLogger(__name__)

def set_absolute_top_price(_min, _top) -> Tuple[float, TopPriceType, float]:
        if not validate_top_min_prices(_top, _min):
            _LOGGER.warning(f"Setting top-price and min-price to zero because of min-price being larger than top-price. Please fix in options. top:{_top} min:{_min}")
            _top = 0
            _min_price = 0
        else:
            _min_price = _min
        if _top is None:
            return float("inf"), TopPriceType.Unset, _min_price
        if _top <= 0:
            return float("inf"), TopPriceType.Unset, _min_price
        return float(_top), TopPriceType.Absolute, _min_price

def validate_top_min_prices(top, min) -> bool:
    if any(
        [top == 0, min == 0]
    ):  
        return True
    return top > min

async def async_validate_top_min_prices(top, min) -> bool:
        if any(
            [top == 0, min == 0]
        ):  
            return True
        return top > min

async def async_add_tomorrow(_day: int, prices_tomorrow: list, absolute_top_price: float) -> float:
        _current_sum = absolute_top_price * _day
        _current_sum += mean(prices_tomorrow)
        ret= _current_sum / (_day + 1)
        #print(absolute_top_price, ret)
        return ret


TOTALPOWER = "Total power"
HOUSEPOWER = "House power"
CARPOWERSENSOR = "CarPower-sensor"
CURRENTPEAKSENSOR = "CurrentPeak-sensor"

AVERAGECONSUMPTION = "Average consumption"
AVERAGECONSUMPTION_24H = "Average consumption 24h"
CONSUMPTION_TOTAL_NAME = "Energy including car"
CHARGERENABLED = "Charger enabled"
CHARGERDONE = "Charger done"